# -*- coding: utf-8 -*-

"""
General description
-------------------

A basic example to show how to model a simple energy system with oemof.solph.

The following energy system is modeled:
                input/output  bgas     bel      bth      
                     |          |        |       |
                     |          |        |       |
 Stadtgas            |--------->|        |       |                    
 Stadtstrom          |------------------>|       |
                     |<------------------|       |
                     |          |        |       |
 BHKW                |------------------>|       |
                     |-------------------------->|
                     |<---------|        |       |
 HeatPump            |<------------------|       |
                     |-------------------------->|
 Heizung & WW        |<--------------------------| 
 Wandheizung         |<--------------------------|
 Hausstrom           |<------------------|       |
                     |------------------>|       |
                     |          |        |       |
 Battery (storage)   |<------------------|       |
                     |------------------>|       |
 Thermstore          |-------------------------->|
                     |<--------------------------|


Data
----
test.csv


Installation requirements
-------------------------

This example requires the version v0.4.x of oemof. Install by:

    pip install 'oemof.solph>=0.4,<0.5'

Optional:

    pip install matplotlib

"""

__copyright__ = "oemof developer group"
__license__ = "GPLv3"

# ****************************************************************************
# ********** PART 1 - Define and optimise the energy system ******************
# ****************************************************************************

###############################################################################
# imports
###############################################################################

# Default logger of oemof
from oemof.tools import logger
from oemof import solph



import logging
import os
import pandas as pd
import pprint as pp

try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = None


solver = "cbc"  # 'glpk', 'gurobi','cbc'...
debug = False #Set number_of_timesteps to 3 to get a readable lp-file.
number_of_time_steps = 8760
solver_verbose = False  # show/hide solver output

# initiate the logger (see the API docs for more information)
logger.define_logging(
    logfile="oemof_example.log",
    screen_level=logging.INFO,
    file_level=logging.DEBUG,
)

logging.info("Initialize the energy system")
date_time_index = pd.date_range(
      "2010/01/01", periods=number_of_time_steps, freq="1H"
)

energysystem = solph.EnergySystem(timeindex=date_time_index)

# Read data file
filename = os.path.join(os.getcwd(), "test.csv")
data = pd.read_csv(filename)
##########################################################################
# Create oemof object
##########################################################################

logging.info("Create oemof objects")

# The bus objects were assigned to variables which makes it easier to connect
# components to these buses (see below).

# create natural gas bus
bgas = solph.Bus(label="natural_gas")

# create electricity bus
bel = solph.Bus(label="electricity")

#wärme
bth = solph.Bus (label="heat")

# adding the buses to the energy system
energysystem.add(bgas, bel, bth)

# create excess component for the electricity bus to allow overproduction
energysystem.add(solph.Sink(label="excess_bel", inputs={bel: solph.Flow(variable_costs=-0.07)}))

# create excess component for the heat bus to allow overproduction
energysystem.add(solph.Sink(label="excess_bth", inputs={bth: solph.Flow(variable_costs=0.01)}))

# create source object representing the natural gas commodity (annual limit)
energysystem.add(
    solph.Source(
        label="Stadtgas",
        outputs={bgas: solph.Flow(full_load_time_max=300000,variable_costs=0.12)},
   )
)

# create fixed source object representing Strom
energysystem.add(
    solph.Source(
        label="Stadtstrom",
        outputs={bel: solph.Flow(full_load_time_max=50000, variable_costs=0.40)},

    )
)

"""
# create fixed source object representing pv power plants
energysystem.add(
    solph.Source(
        label="pv",
        outputs={bel: solph.Flow(fix=data["pv"], nominal_value=582000)},
    )
)
"""

# create simple sink object representing the electrical demand
energysystem.add(
    solph.Sink(
        label="Hausstrom_BB2",
        inputs={bel: solph.Flow(fix=data["h0"],nominal_value=1)},
    )
)

# create simple sink object representing heat demand
energysystem.add(
    solph.Sink(
        label="Heizug&WW_BB2",
        inputs={bth: solph.Flow(fix=data["de_heat"],nominal_value=1)},
    )
)

# create simple sink object representing heat demand
energysystem.add(
    solph.Sink(
        label="Wandheizung_BB2",
        inputs={bth: solph.Flow(fix=data["de_wallheat"],nominal_value=1)},
    )
)
"""
# The cop (coefficient of performance) of the heat pump can be defined as
# a scalar or a sequence.
cop=4
energysystem.add(
    solph.Transformer(
        label='heat_pump',
        inputs={bel: solph.Flow(variable_costs=0.42)},
        outputs={bth: solph.Flow(
           nominal_value=45,
           min=0.0,
           max=1.0
           )},
        conversion_factors={bel:1/cop}
        )
)
"""

#heatpump as offset-Transformer
cop_min = 3       # efficiency at minimal operation point
cop_max = 4       # efficiency at nominal operation point
P_out_min = 30      # absolute minimal output power
P_out_max = 45   # absolute nominal output power

# calculate limits of input power flow
P_in_min = P_out_min / cop_min
P_in_max = P_out_max / cop_max

# calculate coefficients of input-output line equation
c1 = (P_out_max-P_out_min)/(P_in_max-P_in_min)
c0 = P_out_max - c1*P_in_max

# define OffsetTransformer
energysystem.add(
solph.OffsetTransformer(
    label='heatpump',
    inputs={bel: solph.Flow(
        nominal_value=P_in_max,
        max=1,
        min=P_in_min/P_in_max,
        variable_costs=0.42,
        nonconvex=solph.NonConvex())},
     outputs={bth: solph.Flow()},
    coefficients = [c0, c1])
)


"""
#Gasheizung.
#energysystem.add(
#    solph.Transformer(
#        label='gas_burner',
#        inputs={bgas: solph.Flow()},
#        outputs={bth: solph.Flow(min=0,max=1,nominal_value=50,variable_costs=0.01)}
#        )
#)
"""

# create simple BHKW

energysystem.add(
    solph.GenericCHP(
        label='BHKW',
        fuel_input={bgas: solph.Flow(
        H_L_FG_share_max=[0.1 for p in range(0, number_of_time_steps)])},
        electrical_output={bel: solph.Flow(
            P_max_woDH=[11 for p in range(0, number_of_time_steps)],
            P_min_woDH=[7.5 for p in range(0, number_of_time_steps)],
            Eta_el_max_woDH=[0.32 for p in range(0, number_of_time_steps)],
            Eta_el_min_woDH=[0.28 for p in range(0, number_of_time_steps)]
            )},
        heat_output={bth: solph.Flow(
            Q_CW_min=[0 for p in range(0, number_of_time_steps)])},
        Beta=[0.0 for p in range(0, number_of_time_steps)],
        back_pressure=True)
)

"""
#Thermal Storage
energysystem.add(
  solph.GenericStorage(
                    label="Thermstore",
                    inputs={bth: solph.Flow(nominal_value=50)},
                    outputs={bth: solph.Flow(nominal_value=50)},
                    nominal_storage_capacity=200.0,
                    initial_storage_level = 0.0,
                    balanced=False,
                    min_storage_level=0.0,
                    max_storage_level=1,
                    loss_rate=0.0,
                    fixed_losses_relative=0.000,
                    fixed_losses_absolute=0,
                    inflow_conversion_factor=1,
                    outflow_conversion_factor=1,
                )
)

"""

"""
# create storage object representing a battery
energysystem.add(
  solph.GenericStorage(
     label="Battery",
     inputs={bel: solph.Flow(nominal_value=9)},
     outputs={bel: solph.Flow(nominal_value=9)},
     initial_storage_level=0.0, balanced=False,
     loss_rate=0.001, nominal_storage_capacity=12,
     inflow_conversion_factor=0.98, outflow_conversion_factor=0.8)   
)

"""
##########################################################################
# Optimise the energy system and plot the results
##########################################################################

logging.info("Optimise the energy system")

# initialise the operational model
model = solph.Model(energysystem)

# This is for debugging only. It is not(!) necessary to solve the problem and
# should be set to False to save time and disc space in normal use. For
# debugging the timesteps should be set to 3, to increase the readability of
# the lp-file.
if debug:
    filename = os.path.join(
        solph.helpers.extend_basic_path("lp_files"), "basic_example.lp"
    )
    logging.info("Store lp-file in {0}.".format(filename))
    model.write(filename, io_options={"symbolic_solver_labels": True})

# if tee_switch is true solver messages will be displayed
logging.info("Solve the optimization problem")
model.solve(solver=solver, solve_kwargs={"tee": True})

logging.info("Store the energy system with the results.")

# The processing module of the outputlib can be used to extract the results
# from the model transfer them into a homogeneous structured dictionary.

# add results to the energy system to make it possible to store them.
energysystem.results["main"] = solph.processing.results(model)
energysystem.results["meta"] = solph.processing.meta_results(model)

# The default path is the '.oemof' folder in your $HOME directory.
# The default filename is 'es_dump.oemof'.
# You can omit the attributes (as None is the default value) for testing cases.
# You should use unique names/folders for valuable results to avoid
# overwriting.

# store energy system with results
energysystem.dump(dpath=None, filename=None)

# ****************************************************************************
# ********** PART 2 - Processing the results *********************************
# ****************************************************************************

logging.info("**** The script can be divided into two parts here.")
logging.info("Restore the energy system and the results.")
energysystem = solph.EnergySystem()
energysystem.restore(dpath=None, filename=None)

# define an alias for shorter calls below (optional)
results = energysystem.results["main"]

# get all variables of a specific component/bus
#custom_storage = solph.views.node(results, "storage")
electricity_bus = solph.views.node(results, "electricity")
heat_bus = solph.views.node(results, "heat")
gas_bus = solph.views.node(results, "natural_gas")

# plot the time series (sequences) of a specific component/bus
if plt is not None:
    fig, ax = plt.subplots(figsize=(10, 5))
#    custom_storage["sequences"].plot(
#        ax=ax, kind="line", drawstyle="steps-post"
#    )
#    plt.legend(
#        loc="upper center",
#        prop={"size": 8},
#        bbox_to_anchor=(0.5, 1.25),
#        ncol=2,
#    )
#    fig.subplots_adjust(top=0.8)
#    plt.show()

    fig, ax = plt.subplots(figsize=(10, 5))
    electricity_bus["sequences"].plot(
        ax=ax, kind="line", drawstyle="steps-post"
    )
    plt.legend(
        loc="upper center", prop={"size": 8}, bbox_to_anchor=(0.5, 1.3), ncol=2
    )
    fig.subplots_adjust(top=0.8)
    plt.show()
#Plot Thermal    
    fig, ax = plt.subplots(figsize=(10, 5))
    heat_bus["sequences"].plot(
        ax=ax, kind="line", drawstyle="steps-post"
    )
    plt.legend(
        loc="upper center", prop={"size": 8}, bbox_to_anchor=(0.5, 1.3), ncol=2
    )
    fig.subplots_adjust(top=0.8)
    plt.show()

#combine and write results to spreadsheet
output1 = pd.DataFrame.from_dict(heat_bus["sequences"])
output2 = pd.DataFrame.from_dict(electricity_bus["sequences"])
output3 = pd.DataFrame.from_dict(gas_bus["sequences"])
#output=heat_bus
#busmerge=pd.merge(demand, elec_demand_resampled, left_index=True, right_index=True)
with pd.ExcelWriter('output.xlsx') as writer:  
    output1.to_excel(writer, sheet_name="heat_bus")
    output2.to_excel(writer, sheet_name="electricity_bus")
    output3.to_excel(writer, sheet_name="gas_bus")

# print the solver results
print("********* Meta results *********")
pp.pprint(energysystem.results["meta"])
print("")

# print the sums of the flows around the electricity bus
print("********* Main results *********")
print(electricity_bus["sequences"].sum(axis=0))
print(heat_bus["sequences"].sum(axis=0))
print(gas_bus["sequences"].sum(axis=0))